#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <unistd.h>
#include <libaudit.h>
#include <syslog.h>
#include <stdarg.h>

#include "whitelst_db.h"

#define TIME_LEN		60

char *weekday[] = {
	"Sun",
	"Mon",
	"Tue",
	"Wed",
	"Thu",
	"Fri",
	"Sat"
};

void get_local_time(char *tm_local)
{
	time_t timep;
	struct tm *tm;

	time(&timep);
	tm = localtime(&timep);

	sprintf(tm_local, "%d-%02d-%02d %s %02d:%02d:%02d", 
			   tm->tm_year+1900, tm->tm_mon+1, tm->tm_mday, 
			   weekday[tm->tm_wday], tm->tm_hour, tm->tm_min, tm->tm_sec);
}

int audit_record(char *op, char *file, char *hash, int res)
{
	int ret = 0;
	int fa = -1;
	char hostname[65] = {0};
	char message[PATH_MAX] = {0};

	memset(hostname, 0, sizeof(hostname));
	ret = gethostname(hostname, sizeof(hostname));
	if(ret < 0){
		fprintf(stderr, "gethostname failed: %s\n", strerror(errno));
		return -1;
	}

	fa = audit_open();
	if(fa < 0){
		fprintf(stderr, "audit_open failed: %s\n", strerror(errno));
		return -1;
	}

	memset(message, 0, sizeof(message));
	snprintf(message, sizeof(message), "op=\"whitelist:%s\" name=\"%s\" hash=\"%s\"", op, file, hash);

	ret = audit_log_user_message(fa, AUDIT_KYSEC, message, hostname, NULL, NULL, res);
	if(ret <= 0){
		fprintf(stderr, "audit-message failed: %s\n", strerror(errno));
		ret = -1;
	}

	audit_close(fa);

	return ret;
}

int audit_record_for_kysec(char *op, int res)
{
	int ret = 0;
	int fa = -1;
	char hostname[65] = {0};
	char message[PATH_MAX] = {0};

	memset(hostname, 0, sizeof(hostname));
	ret = gethostname(hostname, sizeof(hostname));
	if(ret < 0){
		fprintf(stderr, "gethostname failed: %s\n", strerror(errno));
		return -1;
	}

	fa = audit_open();
	if(fa < 0){
		fprintf(stderr, "audit_open failed: %s\n", strerror(errno));
		return -1;
	}

	memset(message, 0, sizeof(message));
	snprintf(message, sizeof(message), "op=\"whitelist:%s\" name=\"%s\", hash=\"%s\"", op, NULL, NULL);

	ret = audit_log_user_message(fa, AUDIT_KYSEC, message, hostname, NULL, NULL, res);
	if(ret <= 0){
		fprintf(stderr, "audit-message failed: %s\n", strerror(errno));
		ret = -1;
	}

	audit_close(fa);

	return ret;
}


/* syslog */
static void syslog_record(const char *prefix, int level, const char *fmt, va_list ap)
{
	openlog(prefix, LOG_PERROR | LOG_PID | LOG_CONS, LOG_DAEMON);

	vsyslog(level, fmt, ap);

	closelog();
}

void log_debug(const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	syslog_record("KYSEC-WLIST", LOG_DEBUG, fmt, ap);
	va_end(ap);
}

void log_error(const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	syslog_record(NULL, LOG_ERR, fmt, ap);
	va_end(ap);
}

#if 0
int main(void)
{
	char *tm_time = NULL;

	tm_time = malloc(TIME_LEN);
	if(NULL == tm_time){
		fprintf(stderr, "failed to malloc!\n");
		return -1;
	}
	memset(tm_time, 0, TIME_LEN);

	get_local_time(tm_time);
	printf("%s\n", tm_time);

	audit_record("aaa", 1);

	if(tm_time)
		free(tm_time);
	return 0;
}
#endif
