#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <dirent.h>

#include "whitelst_db.h"

#if KYSEC
//#include <kysec/libkysec.h>
//#include <kysec/kmod_protect.h>
#include <kysecwhlist/kysec_whlist.h>
#endif

#define PRO_CONF "/usr/share/kysec-utils/data/readonly_list"

static int file_protect_check(char *file_path)
{
    FILE *fp = NULL;
    char *line = NULL;
    size_t len = 0;
    ssize_t nread;
    int ret = 0;
    char *protect_file = NULL;
    long line_max = 0;
    int wcount = 0;
    struct stat sb;

    if (!file_path)
        return 0;

    fp = fopen(PRO_CONF, "r");
    if (!fp)
    {
        log_error("Failed to open '%s'\n", PRO_CONF);
        return 0;
    }

    line_max = sysconf(_SC_LINE_MAX);
    if (line_max == -1)
    {
        log_error("Cannot get LINE_MAX value\n");
        ret = 0;
        goto out;
    }
    protect_file = malloc(line_max);
    if (!protect_file)
    {
        log_error("Cannot to malloc\n");
        ret = 0;
        goto out;
    }

    while ((nread = getline(&line, &len, fp)) != -1)
    {
        memset(protect_file, 0, line_max);
        memcpy(protect_file, line, nread);

        wcount = ltrim_anyway(protect_file);
        if (wcount == 0 || protect_file[0] == '#')
            continue;

        if ((wcount == 1 && strcmp(protect_file, "/") == 0) ||
            strcmp(file_path, protect_file) == 0)
        {
            ret = 1;
            break;
        }

        if (stat(protect_file, &sb) == -1)
        {
            log_error("Failed to stat with '%s'\n", protect_file);
            continue;
        }
        /* The file from protect-list is a directory */
        if (S_ISDIR(sb.st_mode))
        {
            char *check_file = NULL;
            size_t file_len = 0;
            char *p = NULL;

            if (protect_file[wcount - 1] == '/')
            {
                protect_file[wcount - 1] = '\0';
                wcount--;
            }

            if (strcmp(file_path, protect_file) == 0)
            {
                ret = 1;
                break;
            }

            file_len = strlen(file_path);
            check_file = strndup(file_path, file_len);

            /* The file in protect-list could not be root directory */
            while (file_len > 1)
            {
                if (check_file[file_len - 1] == '/')
                {
                    check_file[file_len - 1] = '\0';
                    file_len--;
                    if (strcmp(check_file, protect_file) == 0)
                    {
                        ret = 1;
                        break;
                    }
                }
                file_len--;
            }

            free(check_file);

            if (ret == 1)
                break;
        }
    }

    if (line)
        free(line);

out:
    fclose(fp);

    if (protect_file)
    {
        free(protect_file);
        protect_file = NULL;
    }

    return ret;
}

/* check if the file is the os held.
 * -1: error, 0: no, 1: yes, 2: not existed
 */
static int check_file_os_simple(char *fpath)
{
    int result = 0;

    if ((access(fpath, F_OK) < 0) && (errno == ENOENT))
    {
        log_debug("The file [%s] is not existed!\n", fpath);
        result = 2;
    }

    return result;
}

int whitelstAdd(int type, char *fpath, int kpro, int os)
{
    int ret = 0;

    switch (type)
    {
    case TYPE_BINARY:
    case TYPE_SCRIPT:
    case TYPE_SHLIB:
    {
        ret = kysec_whlist_add(0, fpath);
        if (ret != 0)
        {
            if (ret > 0)
            {
                if (ret == 6)
                    ret = -3;
                else     
                    ret = -4;
            }
            log_error("Failed to kysec_whlist_add for [%d][%s]!\n", type, fpath);
            goto out;
        }
        break;
    }
    case TYPE_KMOD:
    {
        ret = kysec_whlist_add(2, fpath);
        if (ret != 0)
        {
            if (ret > 0)
            {
                if (ret == 6)
                    ret = -3;
                else     
                    ret = -4;
            }
            log_error("Failed to kysec_whlist_add for [%s]!\n", fpath);
            goto out;
        }
        if (kpro == 1)
        {
            ret = kysec_whlist_update_rmmodpro(1, fpath);
            if (ret != 0)
            {
                ret = -3;
                log_error("Failed to add kmod-protect to kernel for [%s]!\n", fpath);
                goto out;
            }
        }
        break;
    }
    case TYPE_FPROT:
        ret = kysec_whlist_add(1, fpath);
        if (ret != 0)
        {
            if (ret > 0)
            {
                if (ret == 6)
                    ret = -2;
                else     
                    ret = -3;
            }
            log_error("Failed to kysec_protect_set for [%s]!\n", fpath);
            goto out;
        }
        break;
    default:
        log_error("The type[%d] is invalid!\n", type);
        goto out;
    }

out:
    if (ret != 0)
        audit_record("add", fpath, NULL, 0);
    else
        audit_record("add", fpath, NULL, 1);

    return ret;
}

int whitelstDelete(int type, char *fpath)
{
    int ret = -1;
    int check = 0;

    switch (type)
    {
    case TYPE_BINARY:
    case TYPE_SCRIPT:
    case TYPE_SHLIB:
    {
        check = check_file_os_simple(fpath);
        if (check == 2)
        {
            ret = -1;
            goto out;
        }

        ret = kysec_whlist_del(0, fpath);
        if (ret != 0)
        {
            if (ret > 0)
            {
                if (ret == 6)
                    ret = -3;
                else     
                    ret = -4;
            }
            log_error("Failed to kysec_whlist_del for [%d][%s]!\n", type, fpath);
            goto out;
        }
        break;
    }
    case TYPE_KMOD:
    {
        check = check_file_os_simple(fpath);
        if (check == 2)
        {
            ret = -2;
            goto out;
        }

        ret = kysec_whlist_del(2, fpath);
        if (ret != 0)
        {
            if (ret > 0)
            {
                if (ret == 6)
                    ret = -3;
                else     
                    ret = -4;
            }
            log_error("Failed to kysec_whlist_del for [%d][%s]!\n", type, fpath);
            goto out;
        }
        break;
    }
    case TYPE_FPROT:
    {
        // if (file_protect_check(fpath) == 1)
        // {
        //     ret = -5;
        //     log_error("The file '%s' has been protected by system! Could not be removed anymore!\n", fpath);
        //     goto out;
        // }

        ret = kysec_whlist_del(1, fpath);
        if (ret != 0)
        {
            if (ret > 0)
            {
                if (ret == 6)
                    ret = -2;
                else     
                    ret = -3;
            }
            log_error("Failed to kysec_whlist_del for [%d][%s]!\n", type, fpath);
            goto out;
        }
        break;
    }
    default:
        log_error("The type[%d] is invalid!\n", type);
        goto out;
    }

out:
    if (ret != 0)
        audit_record("delete", fpath, NULL, 0);
    else
        audit_record("delete", fpath, NULL, 1);

    return ret;
}

int whitelstAddByDeb(char *debname)
{
    int ret = kysec_whlist_add_by_deb(debname);
    if (ret != 0)
    {
        audit_record("add", debname, NULL, 0);
        return -1;
    }
    else
    {
        audit_record("add", debname, NULL, 1);
        return 0;
    }
}

int whitelstDeleteByDeb(char *debname)
{
    int ret = kysec_whlist_del_by_deb(debname);
    if (ret != 0)
    {
        audit_record("delete", debname, NULL, 0);
        return -1;
    }
    else
    {
        audit_record("delete", debname, NULL, 1);
        return 0;
    }
}

int whitelstAddByDir(int type, char *dpath)
{
    int ret = 0;

    switch (type)
    {
    case TYPE_BINARY:
    {
        ret = kysec_whlist_add_by_dir(0, "program", dpath);
        if (ret != 0)
        {
            if (ret > 0)
            {
                if (ret == 6)
                    ret = -3;
                else     
                    ret = -4;
            }
            log_error("Failed to kysec_whlist_add_by_dir for [%d][%s]!\n", type, dpath);
            goto out;
        }
        break;
    }
    case TYPE_SCRIPT:
    {
        ret = kysec_whlist_add_by_dir(0, "script", dpath);
        if (ret != 0)
        {
            if (ret > 0)
            {
                if (ret == 6)
                    ret = -3;
                else     
                    ret = -4;
            }
            log_error("Failed to kysec_whlist_add_by_dir for [%d][%s]!\n", type, dpath);
            goto out;
        }
        break;
    }
    case TYPE_SHLIB:
    {
        ret = kysec_whlist_add_by_dir(0, "library", dpath);
        if (ret != 0)
        {
            if (ret > 0)
            {
                if (ret == 6)
                    ret = -3;
                else     
                    ret = -4;
            }
            log_error("Failed to kysec_whlist_add_by_dir for [%d][%s]!\n", type, dpath);
            goto out;
        }
        break;
    }
    case TYPE_KMOD:
    {
        ret = kysec_whlist_add_by_dir(2, NULL, dpath);
        if (ret != 0)
        {
            if (ret > 0)
            {
                if (ret == 6)
                    ret = -3;
                else     
                    ret = -4;
            }
            log_error("Failed to kysec_whlist_add_by_dir for [%s]!\n", dpath);
            goto out;
        }
        break;
    }
    case TYPE_FPROT:        
        ret = kysec_whlist_add_by_dir(1, NULL, dpath);
        if (ret != 0)
        {
            if (ret > 0)
            {
                if (ret == 6)
                    ret = -2;
                else     
                    ret = -3;
            }
            log_error("Failed to kysec_whlist_add_by_dir for [%s]!\n", dpath);
            goto out;
        }
        break;
    default:
        log_error("The type[%d] is invalid!\n", type);
        goto out;
    }

out:
    if (ret != 0)
        audit_record("add", dpath, NULL, 0);
    else
        audit_record("add", dpath, NULL, 1);

    return ret;
}

int whitelstDeleteByDir(int type, char *dpath)
{
    int ret = 0;

    switch (type)
    {
    case TYPE_BINARY:
    {
        ret = kysec_whlist_del_by_dir(0, "program", dpath);
        if (ret != 0)
        {
            if (ret > 0)
            {
                if (ret == 6)
                    ret = -3;
                else     
                    ret = -4;
            }
            log_error("Failed to kysec_whlist_del_by_dir for [%d][%s]!\n", type, dpath);
            goto out;
        }
        break;
    }
    case TYPE_SCRIPT:
    {
        ret = kysec_whlist_del_by_dir(0, "script", dpath);
        if (ret != 0)
        {
            if (ret > 0)
            {
                if (ret == 6)
                    ret = -3;
                else     
                    ret = -4;
            }
            log_error("Failed to kysec_whlist_del_by_dir for [%d][%s]!\n", type, dpath);
            goto out;
        }
        break;
    }
    case TYPE_SHLIB:
    {
        ret = kysec_whlist_del_by_dir(0, "library", dpath);
        if (ret != 0)
        {
            if (ret > 0)
            {
                if (ret == 6)
                    ret = -3;
                else     
                    ret = -4;
            }
            log_error("Failed to kysec_whlist_del_by_dir for [%d][%s]!\n", type, dpath);
            goto out;
        }
        break;
    }
    case TYPE_KMOD:
    {
        ret = kysec_whlist_del_by_dir(2, NULL, dpath);
        if (ret != 0)
        {
            if (ret > 0)
            {
                if (ret == 6)
                    ret = -3;
                else     
                    ret = -4;
            }
            log_error("Failed to kysec_whlist_del_by_dir for [%s]!\n", dpath);
            goto out;
        }
        break;
    }
    case TYPE_FPROT:        
        ret = kysec_whlist_del_by_dir(1, NULL, dpath);
        if (ret != 0)
        {
            if (ret > 0)
            {
                if (ret == 6)
                    ret = -2;
                else     
                    ret = -3;
            }
            log_error("Failed to kysec_whlist_del_by_dir for [%s]!\n", dpath);
            goto out;
        }
        break;
    default:
        log_error("The type[%d] is invalid!\n", type);
        goto out;
    }

out:
    if (ret != 0)
        audit_record("delete", dpath, NULL, 0);
    else
        audit_record("delete", dpath, NULL, 1);

    return ret;
}

int getWhitelstCount(int type, char *search)
{
	int ret = -1;

	switch (type)
	{
	case TYPE_BINARY:
	{
		ret = kysec_whlist_get_count_by_obj_type(0, "program", search);
		break;
	}
	case TYPE_SCRIPT:
	{
		ret = kysec_whlist_get_count_by_obj_type(0, "script", search);
		break;
	}
	case TYPE_SHLIB:
	{
		ret = kysec_whlist_get_count_by_obj_type(0, "library", search);
		break;
	}
	case TYPE_KMOD:
	{
		ret = kysec_whlist_get_count_by_obj_type(2, "", search);
		break;
	}
	case TYPE_FPROT:
	{
		ret = kysec_whlist_get_count_by_obj_type(1, "", search);
		break;
	}
	default:
		log_error("The type[%d] is invalid!\n", type);
		goto out;
	}

out:
	if (ret < 0)
		audit_record("getWhitelstCount", NULL, NULL, 0);
	else
		audit_record("getWhitelstCount", NULL, NULL, 1);

	return ret;
}

struct white_list *whitelstShowByType(int type, int num, char *search)
{
	int ret = -1;
	struct object_list list;
	struct object_entry *item = NULL;
	struct white_list *result = NULL;
	int i = 0;
	const unsigned char *fpath;
	const unsigned char *hash;
	size_t fpath_len;
	size_t hash_len;

    // printf("whitelstShowByType type = %d\n", type);
	switch (type)
	{
	case TYPE_BINARY:
	{
		ret = kysec_whlist_load_by_obj_type(0, "program", search, &list);
		break;
	}
	case TYPE_SCRIPT:
	{
		ret = kysec_whlist_load_by_obj_type(0, "script", search, &list);
		break;
	}
	case TYPE_SHLIB:
	{
		ret = kysec_whlist_load_by_obj_type(0, "library", search, &list);
		break;
	}
	case TYPE_KMOD:
	{
		ret = kysec_whlist_load_by_obj_type(2, "kmodule", search, &list);
		break;
	}
	case TYPE_FPROT:
	{
		ret = kysec_whlist_load_by_obj_type(1, "", search, &list);
		break;
	}
	default:
		log_error("The type[%d] is invalid!\n", type);
		goto out;
	}

	if (ret < 0)
	{
		goto err_out;
	}

	result = (struct white_list *)malloc(num * sizeof(struct white_list));
	if (NULL == result)
		goto err_out;

	for (i = 0; i < num; i++)
	{
		memset(result[i].fpath, 0, PATH_MAX);
		memset(result[i].hash, 0, SHA1_SIZE + 1);
	}

	i = 0;
	SLIST_FOREACH(item, &list, entries)
	{
		if (i >= num)
			break;

		fpath = item->object.fpath;
		fpath_len = strlen(fpath);

		strncpy(result[i].fpath, fpath, fpath_len);
		result[i].fpath[fpath_len] = '\0';

		hash = item->object.hash;
		hash_len = strlen(hash);
		strncpy(result[i].hash, hash, hash_len);
		result[i].hash[hash_len] = '\0';

		result[i].kpro = item->object.rmmodpro;
		result[i].os = 0;

		i++;
	}

err_out:
	kysec_whlist_objects_free(&list);

out:
	if (ret < 0)
		audit_record("getWhitelstCount", NULL, NULL, 0);
	else
		audit_record("getWhitelstCount", NULL, NULL, 1);

	return result;
}
